% This script uses MODELESS to determine the modulation recognition
% accuracy with increasing SNR. The default settings employ a 2.5GB
% synthetic dataset, which can be made available upon request. 
% The produced result will resemble Figure 3a from our paper.
% Use the following citation for this code:
% MODELESS: MODulation rEcognition with LimitEd SuperviSion., Wei Xiong, Petko Bogdanov, and Mariya Zheleva., IEEE International Conference on Sensing, Communication and Networking (IEEE SECON 2021).
% Paper available at: http://www.cs.albany.edu/~mariya/lab/papers/modeless_secon21.pdf

% Before running this code, you have to setup the VLFeat library in your
% MATLAB shell. For more information on how to set this up, please visit http://www.vlfeat.org

addpath(genpath('./src'))
addpath(genpath('./vendor'))

assert(exist('vl_fisher')==3, 'Install VL-feat library from http://www.vlfeat.org')

clearvars

rng('default')



DATASET = 'synthetic';

assert(any(strcmp({'synthetic','deepsig'}, DATASET)))



if contains(pwd, '/network/rit/')
  DIRPATH = '/network/rit/lab/zhelevalab/wei_backup';
else
  DIRPATH = '~/Dropbox';
end

addpath(genpath(sprintf('%s/qpas',DIRPATH)))
addpath(genpath(sprintf('%s/libocas_v097',DIRPATH)))


% the modulations being pulled out from dataset
modkeys = {{'BPSK'},{'QPSK'},{'8PSK'},{'16PSK'}, {'QAM16'},{'QAM32'},{'QAM64'},{'QAM128'}};


SUB_DOMAIN = 'emd';

if strcmp(SUB_DOMAIN, 'l2norm')
  S = get_class_similarity_l2norm(cellfun(@(x) x{1}, modkeys,'UniformOutput',false));
elseif strcmp(SUB_DOMAIN, 'emd')
  S = get_class_similarity_emd(cellfun(@(x) x{1}, modkeys,'UniformOutput',false));
else
  assert(false);
end

accr = [];

for SNR = [0:2:20]

fprintf('... process ... snr = %d \n', SNR)

accrAtSNR = [];

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% TRAINING STAGE
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

N_PARTITION = 2;

refS = [1:length(modkeys)];

% generate all unknown classes class partitions
all_refR = pick_n_list(length(modkeys), N_PARTITION);


for ri = [1:size(all_refR,1)]

refR = all_refR(ri,:);

z = length(refR);

N_LEN = 512;

% 每调用一次 从 random stream 取出下一个 随机数
% 所以要保证 在最里一层 loop 里 重置 random state
rngcur = 188;
rng(rngcur,'twister')


if strcmp(DATASET, 'synthetic')
  FIN = sprintf('%s/zsl-preload2/ds_zsl_snr_sample3000_nlen512_seed1817343113-pskqam3.mat', DIRPATH);
elseif strcmp(DATASET, 'deepsig')
  FIN = sprintf('%s/deepsig-preload/ds_zsl_snr_sample3000_nlen1024_trace1.mat', DIRPATH);
else
  assert(false);
end

T = load(FIN,'ds');

X = [];
Y = [];
y = extract_label(T.('ds'), SNR, modkeys);
x = extract_x(T.('ds'), SNR, modkeys);
[x, y, ~] = reduce_dataset_to_num(x, y, 1000*numel(refS));
for i = 1:size(x,1)
  x(i,1:N_LEN) = normalize_signals(x(i,1:N_LEN),4);
end
X = [X; x(:,1:N_LEN)];
Y = [Y; y];

clearvars('x','y');


[X, Y] = compose_zsl_mat(refR, X, Y);


Y_train = Y;

mappedX = [];


N_DICT_SIZE = 30;
N_PATCH_SIZE = 3;

rng(rngcur,'twister')
[Xs, Ys, ~] = reduce_dataset_to_num(X, Y, 200*numel(unique(Y)));

dict = vl_fisher_encoder_train(Xs, Ys, N_PATCH_SIZE, N_DICT_SIZE);
D = {};
D.('dict') = dict;
clearvars('Xs','Ys');

feat = vl_fisher_encoder_encode(D.('dict'), X);
mappedX = [mappedX, feat];



nTrain = size(mappedX,1);
nFeat = size(mappedX,2);
assert(size(mappedX,2) > 0)
X_train = mappedX;
clearvars('mappedX','feat')

X_train = X_train'; % row is feature, column is instance

Templates = zeros(nFeat, length(refR));
for i = 1:length(refR)
  Templates(:,i) = mean(X_train(:,Y_train == refR(i)), 2);
end


H = S(refR,refR);
F = S(refR,:);
alpha = zeros(length(refR), size(F,2));

for i = 1:size(F,2)
  f = -F(:,i);
  alpha(:,i) = qpas(double(H + 1e1*eye(size(H,2))), double(f), [], [], ones(1,length(f)), 1, zeros(length(f),1));
end

% target domain

x = zeros(nFeat, length(refR), length(refR), 'single');
for i = 1:nTrain
  % d = min(repmat(X_train(:,i), [1 size(Templates,2)]), Templates);
  d = max(0, repmat(X_train(:,i), [1 size(Templates,2)])-Templates);
  x(:,:,Y_train(i)==refR) = x(:,:,Y_train(i)==refR) + single(d * alpha(:,refR));
end
y = [];
for i = 1:length(refR)
  d = -ones(size(alpha,2), 1);
  d(refR(i)) = 1;
  y = [y; d(refR)];
end
x = reshape(x, size(x,1), []);

% train svms

SVM_C = 1e1;

maxval = max(abs(x(:)));
[w b stat] = svmocas(x./maxval, 1, double(y), SVM_C);

% update models
for iter = 1:10

  % gradient
  grad = zeros(length(w), size(alpha,1));
  for i = 1:nTrain
    d = max(0, repmat(X_train(:,i), [1, size(Templates,2)])-Templates);
    val = (w' * d) * alpha;
    y = -ones(1, size(alpha,2));
    y(Y_train(i)) = 1;
    decend = single(val.*y < 0);
    if any(decend==1)
      grad = grad + (w*((decend.*y)*alpha')) .* single(repmat(X_train(:,i), [1 size(alpha,1)]) < Templates);
    end
  end
  Templates = max(0, Templates - 1e-3 * grad ./ nTrain);

  % sample data
  x = zeros(nFeat, length(refR), length(refR), 'single');
  for i = 1:nTrain
    % d = min(repmat(X_train(:,i), [1 size(Templates,2)]), Templates);
    d = max(0, repmat(X_train(:,i), [1 size(Templates, 2)]) - Templates);
    x(:,:,Y_train(i)==refR) = x(:,:,Y_train(i)==refR) + single(d*alpha(:,refR));
  end
  y = [];
  for i = 1:length(refR)
    d = -ones(size(alpha,2), 1);
    d(refR(i)) = 1;
    y = [y; d(refR)];
  end
  x = reshape(x, size(x,1), []);

  % train svms
  maxval = max(abs(x(:)));
  [w b stat] = svmocas(x./maxval, 1, double(y(:)), SVM_C);

end

prec_train = (double(nTrain - stat.nTrnErrors))/ double(nTrain);


clearvars('T')

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% INFERENCE STAGE
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

rng(rngcur,'twister')

if strcmp(DATASET, 'synthetic')
  FIN = sprintf('%s/zsl-preload2/ds_zsl_snr_sample3000_nlen512_seed88-pskqam3.mat', DIRPATH);
elseif strcmp(DATASET, 'deepsig')
  FIN = sprintf('%s/deepsig-preload/ds_zsl_snr_sample3000_nlen1024_trace2.mat', DIRPATH);
else
  assert(false);
end


T = load(FIN,'ds');

X = [];
Y = [];
y = extract_label(T.('ds'), SNR, modkeys);
x = extract_x(T.('ds'), SNR, modkeys);
% [x, y, ~] = reduce_dataset_to_num(x, y, 1000*numel(refS));
for i = 1:size(x,1)
  x(i,1:N_LEN) = normalize_signals(x(i,1:N_LEN),4);
end
X = [X; x(:,1:N_LEN)];
Y = [Y; y];
clearvars('x','y');


refT = setdiff(refS, refR);

z = length(refT);

[X, Y] = compose_zsl_mat(refT, X, Y);


Y_test = Y;


mappedX = vl_fisher_encoder_encode(D.('dict'), X);


assert(size(mappedX,2) > 0)
X_test = mappedX;
clearvars('mappedX','feat')
nTest = size(X_test,1);
assert(size(X_test,2) == nFeat);

X_test = X_test'; % row is feature, column is instance


margins = [];
for i = 1:nTest
  % d = min(repmat(X_test(:,i), [1 size(Templates,2)]), Templates);
  d = max(0, repmat(X_test(:,i), [1 size(Templates,2)]) - Templates);
  d = w' * d * alpha(:, refT);
  margins = [margins; d];
end

[~, id] = max(margins, [], 2);
prec = sum(refT(id)'==Y_test)/nTest;

fprintf('[test set] train %s test %s accuracy %.4f \n', sprintf('%d ',refR),sprintf('%d ',refT), prec);


accrAtSNR = [accrAtSNR prec];

clearvars('T','D')
clearvars('w','b','stat')

end % / for refR

accr = [accr, mean(accrAtSNR)];

end % ./ for SNR

accr